# -*- coding: utf-8 -*-
"""Crypto-Factory

Main purpose of this library is to provide a common interface to application
for cryptographic tasks.
"""

# TODO: Update docstrings.
# TODO: Finish writing documentation
# TODO: Create HOWTOs
# TODO: Import providers only if required pkg(s) present
# TODO: {Last} Clean up code !!!


from .interface import CryptoFactory
from . import templates
from . import providers
from . import utils
from . import exceptions


__all__ = [
    'CryptoFactory',
    'templates',
    'providers',
    'utils',
    'exceptions',
]

name = "crypto_factory"

__title__ = "Crypto-Factory"
__summary__ = "Standardize cryptographic tasks in your application."
__uri__ = "https://github.com/LuckyDams/Crypto-Factory"

__version__ = "0.0.9"
__version_info__ = tuple(int(x) for x in __version__.split('.'))
__build__ = "20190618"

__author__ = "LuckyDams"
__email__ = "LuckyDams@gmail.org"

__license__ = "MIT"
__copyright__ = "Copyright 2019 {}".format(__author__)




# -*- coding: utf-8 -*-
"""
utilities
---------

This module provides the ``CryptoUtils`` class available within the Crypto-Factory
interface through the ``utils`` property.
"""


__all__ = [
    'CryptoUtils',
]


import os
import base64
import string
import secrets
import uuid
from .exceptions import CryptoFactoryError


class CryptoUtils:
    valid_tag_chars = "!#(),-.:;<>?@[]_{|}%s%s" % (string.ascii_letters, string.digits)

    @classmethod
    def normalize_tag(cls, tag):
        """
        .. _normalize-tag:

        **Normalize tag** value

        If tag is type *int*, will return its string representation (*str*).

        If tag is type *dict*, will look for 'tag' key value ; then will
        validate string  as *str*.

        If tag is type *str*, will check if its characters are valid then return
        a normalized tag in upper format:

            - Valid characters are ascii letters, digits and punctuation in
              "``!#(),-.:;<>?@[]_{|}``".

            - "`$`" (used as separator), and other characters will be removed.

        :param tag: Value to normalize as 'tag'.
        :type tag: str or int or dict

        :return: **normalized tag** (upper `str`) OR `None`
        """

        if isinstance(tag, int):
            return str(tag)

        if isinstance(tag, dict):
            tag = tag.get('tag', '')

        if isinstance(tag, str):
            tag = ''.join(c for c in tag if c in cls.valid_tag_chars)
            if len(tag):
                return tag.upper()
            else:
                raise CryptoFactoryError('Null or invalid tag')

        else:
            raise CryptoFactoryError('Invalid tag type')

    @classmethod
    def get_tag(cls, value):
        """**Get tag** from string (encrypted data)"""
        value = cls.data_string(value)
        return value[:value.find('$')]

    @classmethod
    def remove_tag(cls, value):
        """**Remove tag** from string (encrypted data)"""
        value = cls.data_string(value)
        pos = value.find('$') + 1
        return value[pos:]

    @classmethod
    def compare_tag(cls, value, tag):
        """**Compare tag**.
        Return `True` if tag in string (encrypted data), else `False`.
        """
        if tag == cls.get_tag(value):
            return True

    @staticmethod
    def generate_key(size=256, urlsafe=True):
        """
        Generate **random keys** (as *bytes*) of 128, 192, 256, or 512 bits size.
        (standard or urlsafe ``base64`` encoded.)
        """
        if size in (128, 192, 256, 512):
            rand = os.urandom(size // 8)
            if urlsafe:
                # either are working :-)
                return base64.urlsafe_b64encode(rand)
                # return ''.join((secrets.token_urlsafe(size // 8), '=')).encode()
            else:
                return base64.b64encode(rand)
        else:
            raise CryptoFactoryError('Invalid key size')

    @staticmethod
    def generate_tag():
        """Generate a **unique tag** (return a standard ``uuid4`` as *str*)"""
        return str(uuid.uuid4())

    @classmethod
    def generate_string(cls, length=8, chars=None, tag=False):
        """
        Generate a **random string** value of 'n' valid characters.

        :param int length:
            Number of characters to return.

        :param str chars:
            Valid characters tu use. If *None*, will use any charaters from
            ascii_letters, digits & punctuation (from ``string`` library).

        :param bool tag:
            If *True*, will use valid characters for tag.
            (take precedence on `chars` argument)

        :return: str
            (None if length is not int)
        """
        if isinstance(length, int):
            valid_chars = "%s%s%s" % (string.ascii_letters, string.digits, string.punctuation)

            if chars and isinstance(chars, str):
                valid_chars = chars

            if tag:
                valid_chars = cls.valid_tag_chars

            return ''.join(secrets.choice(valid_chars) for i in range(length))

    @staticmethod
    def data_string(data, encoding='utf-8', errors='strict'):
        """Convert **data** to `str` using  standard encoding values if data is `bytes`."""
        return data.decode(encoding, errors) if isinstance(data, bytes) else data

    @staticmethod
    def data_bytes(data, encoding='utf-8', errors='strict'):
        """Convert **data** to `bytes` using  standard encoding values if data is `str`."""
        return data.encode(encoding, errors) if isinstance(data, str) else data

    @staticmethod
    def b64encode(s, urlsafe=True):
        if urlsafe:
            return base64.urlsafe_b64encode(s)
        else:
            return base64.b64encode(s)

    @staticmethod
    def b64decode(s, urlsafe=True):
        try:
            if urlsafe:
                return base64.urlsafe_b64decode(s)
            else:
                return base64.b64decode(s)

        except base64.binascii.Error as err:
            raise CryptoFactoryError(err)



# -*- coding: utf-8 -*-
"""
interface
---------

This module owns the client interface within the main class ``CryptoFactory``.
"""

# TODO: Check Crypto service 'features' to ensure method is implemented


__all__ = [
    'CryptoFactory',
]


from .factory import CryptoServicesManager
from .utils import CryptoUtils
from .exceptions import CryptoFactoryError
from .templates import CryptoBuilder, CryptoService


class CryptoFactory:
    """
    Client **interface** for Crypto-Factory.

    It provides following features access:

        - **Crypto services management** (See ``factory`` module.)

        - Abstraction **methods for encryption**; rely on underneath Crypto
          services (See ``providers`` module.)

        - **Tagging** mechanism

        - **Advanced method(s)** such as ``rotate`` to migrate encrypted data
          to another Crypto service

        - **Crypto utilities** such as generate random keys and strings, inspect
          tags or encoding functions (See ``utilities`` module.)

        - **Built-in Crypto providers** for direct use of defined encryption
          algorithms (See ``providers`` module.)

        - **Crypto templates** for building your own encryption algorithms
          (See ``providers`` module.)
    """
    def __init__(self):
        self.__services = CryptoServicesManager()
        self.__utils = CryptoUtils
        self.CryptoBuilder = CryptoBuilder
        self.CryptoService = CryptoService
        # self._caller = _get_caller()

    @property
    def services(self):
        """entry-point to **Crypto services manager**
        (See ``CryptoServicesManager`` class for details.)
        """
        return self.__services

    @property
    def utils(self):
        """entry-point to **Crypto utilities**
        (See ``CryptoUtils`` class for details.)
        """
        return self.__utils

    def __add_tag(self, value, mode):
        """Add tag to string"""
        tag = self.__services.get_mode_tag(mode)
        if tag:
            return '$'.join((tag, value))

    def __compare_tag(self, value, mode):
        """Compare tag from string to defined Crypto service tag"""
        tag = self.__services.get_mode_tag(mode)
        return self.__utils.compare_tag(value, tag)

    def __del_tag(self, value):
        """Remove tag from string"""
        return self.__utils.remove_tag(value)

    def encrypt(self, data, mode, tag=True):
        """
        Abstraction method for **symmetric encryption**.

        Call ``encrypt`` method of selected Crypto service.
        Will add a 'tag' to encrypted 'data' before returning result if ``tag=True``.

        :param data:
            Plain data (to encrypt).
        :type data: str or bytes

        :param mode:
            Crypto service identifier ('sid' in ``factory``).
        :type mode: str or int

        :param bool tag:
            Add the Crypto service tag to 'data' if *True* (default).
            If Crypto service has tagging disabled (no 'tag' on ``register``),
            this parameter will be skipped (set to *False*).

        :return: **encrypted data** as *str*

        :raises CryptoFactoryError:
            An exception is raised if an error occurs. Check error message for details.
        """
        # Check args
        if not isinstance(data, (str, bytes)):
            raise CryptoFactoryError('Supported data format is string or bytes')
        # if tag and not self.__services.get_mode_tag(mode):
        #     raise CryptoFactoryError('No defined tag for this Crypto service (tag=True)')

        # Disable tag if not defined in provider
        if not self.__services.get_mode_tag(mode):
            tag = False

        # data as str
        data = self.__utils.data_string(data)

        mode = self.__services.get_mode(mode)

        if data:
            try:
                enc_data = self.__services.get(mode).encrypt(data)
                enc_data = self.__utils.data_string(enc_data)
            except:
                raise CryptoFactoryError("Unable to encrypt with Crypto service: " + str(mode))

            if tag:
                enc_data = self.__add_tag(enc_data, mode)

            return enc_data

    def decrypt(self, data, mode=None, tag_strict=False):
        """
        Abstraction method for **symmetric decryption**.

        Check if tag is matching selected Crypto service 'mode' (if 'tag_strict' enabled).
        Remove tag before calling  ``decrypt`` method of selected Crypto service.

        :param data:
            Encrypted data (to decrypt).
        :type data: str or bytes

        :param mode:
            Crypto service identifier ('sid' in ``factory``).
        :type mode: str or int

        :param bool tag_strict:
            Enforce tag validation (*False* by default).
            If *True*, compare data tag with Crypto service tag.

        :return: **decrypted data** as *str*

        :raises CryptoFactoryError:
            An exception is raised if an error occurs. Check error message for details.
        """
        # Check args
        if not isinstance(data, (str, bytes)):
            raise CryptoFactoryError('Supported data format is string or bytes')
        if tag_strict:
            if mode is None:
                raise CryptoFactoryError('Crypto service mode required (tag_strict enabled)')
            elif not self.__services.get_mode_tag(mode):
                raise CryptoFactoryError('No tag defined for this Crypto service (tag_strict enabled)')
        # data as str
        data = self.__utils.data_string(data)

        # Define mode
        if mode is None:
            # retrieve from tag if not provided
            tag = self.__utils.get_tag(data)
            mode = self.__services.get_tag_mode(tag)
        mode = self.__services.get_mode(mode)

        # Check matching tag in data
        if tag_strict:
            if not self.__compare_tag(data, mode):
                raise CryptoFactoryError('Tag mismatch found (tag_strict enable)')

        # Remove tag (if any)
        data = self.__del_tag(data)

        if data:
            try:
                plain_data = self.__services.get(mode).decrypt(data)
            except:
                raise CryptoFactoryError("Unable to decrypt with Crypto service: " + str(mode))

            return self.__utils.data_string(plain_data)

    def rotate(self, data=None, from_modes=None, to_mode=None, tag_strict=False, tag=True):
        """
        Advanced method to **'rotate' encrypted data** from one or more Crypto
        services to a target one. It can be used for keys rotation, migration of
        existing data to a new Crypto service, ...

        Based on decrypt and encrypt methods, will provide same tag options.

        :param data:
            Encrypted data (to rotate).
        :type data: str or bytes

        :param from_modes:
            Source Crypto service identifier(s).
        :type from_modes: str or list

        :param to_mode:
            Target Crypto service identifier.
        :type to_mode: str or int

        :param bool tag_strict:
            Enforce tag validation (see ``decrypt`` method).

        :param bool tag:
            Add the Crypto service tag (see ``encrypt`` method).

        :return: **encrypted data** as *str* (to target Crypto service format)

        :raises CryptoFactoryError:
            An exception is raised if an error occurs. In particular when
            no 'from_modes' methods are able to decrypt data.
            Check error message for details.
        """
        if data:
            modes_list = []
            if isinstance(from_modes, (str, int)):
                modes_list.append(from_modes)
            elif isinstance(from_modes, (list, tuple)):
                modes_list = from_modes

            for mode in modes_list:
                try:
                    tmp_data = self.decrypt(data, mode, tag_strict)
                    break
                except CryptoFactoryError:
                    pass
            else:
                raise CryptoFactoryError("Unable to rotate data with provided list of Crypto services")

            enc_data = self.encrypt(tmp_data, to_mode, tag)
            return enc_data


if __name__ == '__main__':
    None



# -*- coding: utf-8 -*-
"""
providers
---------

This module contains **built-in Crypto builders** and associated services.
"""

# TODO: Implement new built-in providers: full AES, PBKDF2, MultiFernet, ...


__all__ = [
    'DummyServiceBuilder',
    'FernetServiceBuilder',
    'AESServiceBuilder',
]


import base64
# import inspect
import string
from secrets import token_bytes, choice

from cryptography.fernet import Fernet, InvalidToken
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import padding
from cryptography.hazmat.primitives.ciphers import Cipher, algorithms, modes

from .templates import CryptoService, CryptoBuilder
from .utils import CryptoUtils
from .exceptions import CryptoFactoryError


class DummyService(CryptoService):
    """Dummy Service"""

    def __init__(self, cipher):
        super().__init__(cipher)
        self.pre_label = '<' + cipher + '>'
        self.post_label = '</' + cipher + '>'

    def encrypt(self, raw):
        raw = CryptoUtils.data_string(raw)

        if isinstance(raw, str) and len(raw) > 0:
            enc = ''.join((self.pre_label, raw, self.post_label))
            return enc

    def decrypt(self, enc):
        enc = CryptoUtils.data_string(enc)

        if isinstance(enc, str):

            if len(enc) == 0:
                raise ValueError("Empty input string")

            try:
                out = enc.split(self.pre_label, 1)[1].rsplit(self.post_label, 1)[0]
                return out

            except IndexError:
                raise ValueError("Invalid key or string")


class DummyServiceBuilder(CryptoBuilder):
    """
    Builder for ``DummyService``:
        For illustration purpose only (do NOT use it in production !!!)

    specifications:
        - Surround data with ``label``
    """
    service = DummyService

    @staticmethod
    def initialize(label=None):
        if label:
            return label.lower()


class FernetService(CryptoService):
    """Fernet Service"""

    def encrypt(self, raw):
        raw = CryptoUtils.data_bytes(raw)

        if isinstance(raw, bytes) and len(raw) > 0:
            enc = self.cipher.encrypt(raw)
            return enc.decode('utf-8')

    def decrypt(self, enc):
        enc = CryptoUtils.data_bytes(enc)

        if len(enc) == 0:
            raise ValueError("Empty input string")

        try:
            out = self.cipher.decrypt(enc)
            return out.decode('utf-8')

        except InvalidToken:
            raise ValueError("Invalid key or string")


class FernetServiceBuilder(CryptoBuilder):
    """
    Builder for ``FernetService``:
        Fernet implementation (secured)

    specifications:
        - AES / CBC mode / PKCS7 padding
        - HMAC authentication with SHA256
        - 256 bits ``key`` (used for AES & HMAC)
        - random 128-bit IV
        - urlsafe Base64 encoding
        - string to bytes conversion in UTF-8

    (https://cryptography.io/en/latest/fernet/)
    """
    service = FernetService

    @staticmethod
    def initialize(key=None):
        if key:
            try:
                cipher = Fernet(key)
                return cipher

            except (base64.binascii.Error, ValueError):
                raise ValueError("Invalid key")


class AESService(CryptoService):
    """AESService"""

    def __init__(self, kwargs):
        self.key = kwargs.get('key')
        self.iv = kwargs.get('iv')
        self.prefix = kwargs.get('prefix')
        self.salt = kwargs.get('salt')
        self.urlsafe = kwargs.get('urlsafe')
        self.encoding = kwargs.get('encoding')

    def encrypt(self, data):
        prefix = self.prefix
        # salt = ''.join(choice(string.printable) for i in range(self.salt))
        salt = CryptoUtils.generate_string(self.salt, chars=string.printable)
        data = ''.join((prefix, salt, data))

        data = CryptoUtils.data_bytes(data, self.encoding)

        if len(data) > 0:
            iv = self.iv or token_bytes(16)
            f = Cipher(algorithms.AES(self.key), modes.CBC(iv), backend=default_backend())
            encryptor = f.encryptor()
            padder = padding.PKCS7(128).padder()
            padded_data = padder.update(data)
            padded_data += padder.finalize()
            ct = encryptor.update(padded_data) + encryptor.finalize()
            if self.iv:
                data = ct
            else:
                data = iv + ct
            enc = CryptoUtils.b64encode(data, self.urlsafe)
            return enc

    def decrypt(self, enc):
        pos = len(self.prefix) + self.salt

        if len(enc) > pos + 16:
            try:
                enc = CryptoUtils.b64decode(enc, self.urlsafe)
                if self.iv:
                    iv = self.iv
                else:
                    iv = enc[:16]
                    enc = enc[16:]
                f = Cipher(algorithms.AES(self.key), modes.CBC(iv), backend=default_backend())
                decryptor = f.decryptor()
                unpadder = padding.PKCS7(128).unpadder()
                ct = decryptor.update(enc) + decryptor.finalize()
                unpadded_data = unpadder.update(ct)
                unpadded_data += unpadder.finalize()
                data = unpadded_data.decode(self.encoding, 'ignore')

                if data.startswith(self.prefix):
                    return data[pos:]

            except ValueError:
                pass

        raise CryptoFactoryError("Invalid key or string")


class AESServiceBuilder(CryptoBuilder):
    """
    Builder for ``AESService``:
        AES-CBC-PKCS7 implementation

    specifications:
        - AES / CBC mode / PKCS7 padding
        - key of different sizes (128, 192 or 256 bits)
        - random of fixed IV (16 bytes) support
        - optional prefix (fixed string)
        - optional random salt addition (``salt_length`` in bytes)
        - standard or urlsafe base64 encoding
        - standard string encodings support

    string format:
        ``<random_IV: 16><prefix><salt><data>``
    """
    service = AESService
    # _defaults = {}
    _defaults = dict(
        iv=None,
        prefix=None,
        salt_length=None,
        urlsafe=None,
        encoding=None)

    @classmethod
    def _get_defaults(cls, value):
        return cls._defaults.get(value)

    def initialize(self, key, iv=None, prefix=None, salt_length=None, urlsafe=None, encoding=None):
        """
        AESServiceBuilder initialisation

        :param bytes key:
            key of 16, 24 or 32 bytes

        :param bytes iv:
            fixed IV of 16 bytes
            If None, a random IV will be generated on each call and added to
            encrypted data.

        :param str prefix:
            Optional string of data added to encrypted data.

        :param int salt_length:
            Optional random value of 'salt_length' bytes added to encrypted data.

        :param bool urlsafe:
            If True, use 'urlsafe' Base64 encoding. Else, use standard Base64.
            This apply to both key and iv values. Ensure same Base64 method has
            been used to generate them.

        :param str encoding:
            Any supported Python encoding to use for Str / Bytes conversion.
            (See `codecs <https://docs.python.org/3/library/codecs.html#
            standard-encodings>`_ for details)

        :return: dict of parameters for ``AESService``

        .. _provider-subclassing:
        .. Note::
            This provider can be sub-classed in order to define parameters
            default values. This would allow to design a specific AES
            implementation without the need to provide all parameters on each
            initialisation.

            These parameters can bet set in the ``_defaults`` class attribute
            (dict). ``key`` parameter cannot be defined as defaults as it should
            always be provided from client.

            example::

                class MyAESBuilder(AESServiceBuilder):
                    _defaults = dict(
                        iv= None,
                        prefix='MyAES-prefix',
                        salt_length=24,
                        urlsafe=False,
                        encoding='utf-16',
                    )

        """
        # get args or defaults (helper for subclassing)
        iv = iv or self._get_defaults('iv')
        prefix = prefix or self._get_defaults('prefix')
        salt_length = salt_length or self._get_defaults('salt_length')
        urlsafe = urlsafe if isinstance(urlsafe, bool) else self._get_defaults('urlsafe')
        encoding = encoding or self._get_defaults('encoding')

        # validate values
        if not isinstance(prefix, str):
            prefix = ''

        if not isinstance(salt_length, int):
            salt_length = 0

        if not isinstance(urlsafe, bool):
            urlsafe = True

        if not encoding:
            encoding = 'utf-8'

        if iv:
            iv = CryptoUtils.b64decode(iv, urlsafe)
            if len(iv) != 16:
                raise CryptoFactoryError('Invalid IV length (not 16 bytes)')

        # Init cipher if key is valid
        key = CryptoUtils.b64decode(key, urlsafe)
        if len(key) in (16, 24, 32):
            cipher = dict(key=key, iv=iv, prefix=prefix, salt=salt_length, urlsafe=urlsafe, encoding=encoding)
            return cipher
        else:
            raise CryptoFactoryError('Invalid key length')


if __name__ == '__main__':
    None



import pathlib
from setuptools import setup, find_packages

import crypto_factory as cf

# The directory containing this file
HERE = pathlib.Path(__file__).parent

# The text of the README.md file
README = (HERE / "README.md").read_text()

# Requirements list
REQUIRED_PKGS = (HERE / "requirements.txt").read_text().split()

# This call to setup() does all the work
setup(
    name=cf.__title__,
    version=cf.__version__,
    author=cf.__author__,
    author_email=cf.__email__,
    description=cf.__summary__,
    long_description=README,
    long_description_content_type="text/markdown",
    # long_description_content_type="text/x-rst",
    url=cf.__uri__,
    license=cf.__license__,
    packages=find_packages(exclude=("tests", "private", )),
    include_package_data=True,
    platforms='any',
    # zip_safe=False,
    python_requires='>=3.4',
    install_requires=REQUIRED_PKGS,
    test_suite="tests",
    keywords='cryptography factory-design',
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Topic :: Security :: Cryptography",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
    project_urls={
        'Documentation': 'https://' + cf.__title__.lower() + '.readthedocs.io/en/latest/',
        'Releases': 'https://pypi.org/project/' + cf.__title__,
        'Source': cf.__uri__,
        'Tracker': cf.__uri__ + '/issues',
    },
)



# -*- coding: utf-8 -*-
"""
exceptions
----------

This module provides the ``CryptoFactoryError`` Exception class.
"""

__all__ = [
    'CryptoFactoryError',
]


class CryptoFactoryError(Exception):
    """
    Crypto-Factory exceptions class to simplify errors management in client interface.
    """
    pass



# -*- coding: utf-8 -*-
"""
factory
-------

This module is responsible of managing Crypto providers through the ``CryptoServicesManager`` class.
"""

# TODO: Need solution to import external Builder by name
# TODO: Create a dedicated class to manage tags
# TODO: Create a dedicated class to manage Builders index
# TODO: Create a class to detect built-in and user-defined Builders


__all__ = [
    'CryptoServicesManager',
]


import inspect
from attr_dict import LazyIndex
from .templates import CryptoBuilder
from .utils import CryptoUtils
from .exceptions import CryptoFactoryError


# class CryptoTagsUtility:
#     pass


# class CryptoIndex:
#     """Index for Crypto Services"""
#
#     def __init__(self):
#         self.__providers = LazyIndex()
#         self.__tags = LazyIndex()
#
#     @property
#     def providers(self):
#         return self.__providers.index
#
#     @property
#     def tags(self):
#         return self.__tags.index
#
#     def get_mode(self, sid):
#         """Get Crypto service instance by its identifier"""
#         if sid in self.providers:
#             return sid
#         raise CryptoFactoryError("Unknown or unavailable mode")
#
#     def get_mode_tag(self, sid):
#         """Get Crypto service tag by its identifier"""
#         return self.__providers.get_arg(sid, 'tag')
#
#     def get_tag_mode(self, tag):
#         """Get Crypto service identifier by its tag"""
#         return self.__tags.get_arg(tag, 'id')


# class DefinedBuilders:
#     """
#     DefinedBuilders:
#         Helper class to detect builders builtins in 'providers' module.
#
#     """
#
#     # _locals = [x for x in globals().keys() if x.endswith('Builder')]
#     # locals = {
#     #     'DummyServiceBuilder': DummyServiceBuilder(),
#     #     'AES1ServiceBuilder': AES1ServiceBuilder(),
#     #     'FernetServiceBuilder': FernetServiceBuilder(),
#     # }
#
#     def __init__(self):
#         self.builtins = self.get_builtins
#         self.locals = self.get_locals
#
#     @property
#     def get_builtins(self):
#         import crypto_factory.providers
#         builders = self.__get_builders(crypto_factory.providers)
#         return builders
#
#     @property
#     def get_locals(self):
#         # FIXME: get local builders (sys.modules[__name__] method is not working)
#         import sys
#         builders = self.__get_builders(sys.modules[__name__])
#         return builders
#
#     @staticmethod
#     def __get_builders(module):
#         builders = dict([m for m in inspect.getmembers(module, inspect.isclass)
#                          if m[1].__base__.__name__ == CryptoBuilder.__name__])
#         return builders


# def _get_caller():
#     stack = inspect.stack()
#     module = inspect.getmodule(stack[2][0])
#     return module


def _get_builtins():
    """Function to provide built-in Crypto services builders as a dict of <name>:<class>"""
    import crypto_factory.providers
    builders = dict([
        m for m in inspect.getmembers(crypto_factory.providers, inspect.isclass)
        if m[1].__base__.__name__ == CryptoBuilder.__name__
    ])
    return builders


class CryptoServicesManager:
    """
    CryptoServiceManager is the "factory" class which will manage the Crypto
    providers and will handle calls from the client interface.
    """
    def __init__(self):
        # Create entry-points using LazyIndex
        self.__providers = LazyIndex()
        self.__tags = LazyIndex()
        self.__builtins = LazyIndex()
        builtins = _get_builtins()
        for k, v in builtins.items():
            self.__builtins.set_key(k, v)

        # self._locals = DefinedBuilders().locals
        # self._builtins = DefinedBuilders().builtins
        self.__configs = {}

    @property
    def providers(self):
        """Index of active Crypto services"""
        return self.__providers.index

    @property
    def tags(self):
        """Index of current tags associated to a Crypto services"""
        return self.__tags.index

    @property
    def _builtins(self):
        """Index of available built-in Crypto providers"""
        return self.__builtins.index

    def get_mode(self, sid):
        """Get Crypto service instance by its identifier"""
        if sid in self.providers:
            return sid
        raise CryptoFactoryError("Unknown or unavailable mode")

    def get_mode_tag(self, sid):
        """Get Crypto service tag by its identifier"""
        return self.__providers.get_arg(sid, 'tag')

    def get_tag_mode(self, tag):
        """Get Crypto service identifier by its tag"""
        return self.__tags.get_arg(tag, 'id')

    def register(self, config=None, sid=None, builder=None, tag=None):
        """
        Register and initialize a Crypto service using its builder class.

        :param dict config:
            Crypto service parameters to initialize it (dict).
            Required keys/values are dependant to arguments needed by
            the Crypto service (ie: key, iv, ...). Can also have ``id``,
            ``builder`` and ``tag`` keys to replace following arguments.

        :param sid:
            Crypto service identifier to register with.
            (Equivalent to ``mode`` argument in client interface)
        :type sid: str or int

        :param builder:
            Crypto service builder as a child class of ``CryptoBuilder``.
            Class name as *str* also supported for built-in providers.

            .. Note::
                If ``builder=None``, a null entry will be created for this
                service identifier. This is the method to use to deregister
                a service.

        :type builder: class or str

        :param str tag:
            Crypto service tag as a string of characters (tag can be used as
            identifier on encrypted data). The tag will be formatted using
            ``normalize_tag`` method (from ``CryptoUtils`` class). "``$``"
            character is reserved as it is used as separator.
            (See :ref:`API <normalize-tag>` for details)

        :return: ``sid`` or ``None``
            Return ``sid`` if builder has been successfully registered and
            initialized as an available service.
            Return ``None`` if builder has not been initialized.

        :raises CryptoFactoryError:
            If the Crypto service is unable to register or initialize, an error
            is raised. Check error message for details.


        """

        # Check config
        if config and isinstance(config, dict):
            if sid is None:
                sid = config.get('sid') or config.get('id')
            builder = builder or config.get('builder')
            tag = tag or config.get('tag')

        if isinstance(sid, (str, int)):

            # Define builder
            if isinstance(builder, str):
                if self._builtins.get(builder):
                    builder = self._builtins[builder]
                else:
                    # builder = globals().get(builder)  # not working
                    raise CryptoFactoryError('Not a built-in Crypto builder')

            if isinstance(builder, type):
                # issubclass not working with external module, need alternate solution <class>.__base__ ?= CryptoBuilder
                # if issubclass(builder, CryptoBuilder):

                # if builder.__base__.__name__ == CryptoBuilder.__name__:
                # New check to support sub-child inheritance (templating an existing Builder)
                # if builder.__weakref__.__objclass__ == CryptoBuilder:
                if CryptoBuilder in builder.__mro__:
                    builder = builder()
            else:
                builder = None

            # Define tag
            tag = CryptoUtils.normalize_tag(tag) if tag else None

            # Check if tag already registered with a service id
            if self.get_tag_mode(tag) is not None:
                if builder:
                    # We are trying to register a service with an existing tag !!!
                    raise CryptoFactoryError('Tag already used by a registered provider')

            # Deregister a service when builder is None
            if not builder:
                # Remove tag entry before service
                registered_tag = self.get_mode_tag(sid)
                self.__tags.del_key(registered_tag)
                # Create a None entry in providers
                self.__providers.set_arg(sid, id=sid, builder=None, tag=None)

            # Register service
            else:
                # Record service in providers index
                self.__providers.set_arg(sid, id=sid, builder=builder, tag=tag)
                # Update tags index if tag
                if tag:
                    self.__tags.set_arg(tag, id=sid)
                # Record config (UNSAFE: Uncomment only for debug.)
                # self.__configs[sid] = config

            # Init service
            if builder and isinstance(config, dict):

                try:
                    builder(**config)
                    return sid
                except:
                    raise CryptoFactoryError('Unable to start provider')

        else:
            raise CryptoFactoryError('Unable to register provider, missing or wrong type argument(s)')

    def get(self, sid):
        """
        Call to a Crypto service instance

        This method is called by abstraction methods from the client interface
        and should not be used directly.

        :param sid:
            Crypto service identifier to call.
            (Equivalent to ``mode`` argument in client interface)
        :type sid: str or int

        :return: **Crypto service builder** class instance

        :raise CryptoFactoryError:
            If the Crypto service is not registered.
        """
        builder = self.__providers.get_arg(sid, 'builder')
        if builder:
            return builder()
        else:
            raise CryptoFactoryError('Service not registered: ', sid)


if __name__ == '__main__':
    None



# -*- coding: utf-8 -*-
"""
templates
---------

This module contains ``CryptoBuilder`` & ``CryptoService`` parent classes to use when designing a Crypto provider.
"""

# TODO: Update docstrings
# TODO: Rework templating to improve user implementation


__all__ = [
    'CryptoBuilder',
    'CryptoService',
]


import inspect


class CryptoBuilder:
    """
    Template to design a **Crypto Builder**. Main purpose is to define an
    ``initialize`` method to assign to a Crypto Service.

    This Crypto Service must be designated in the ``service`` class attribute.

    By default, the ``initialize`` method will forward all parameters received by
    the ``register`` command as a dict. This method can be personalised to fulfill
    the needs of the implemented cryptography recipe.

    All defined Crypto Builders must inherit from the ``CryptoBuilder`` class.

    Example::

        class MyBuilder(CryptoBuilder):
            service = MyService

            @staticmethod
            def initialize(key=None):
                ...
                cipher = MyCryptoRecipe(key)
                return cipher

    Note:
        See this **warning** on :ref:`sensitive data in memory <memory-warning>`.
    """
    service = None

    def __init__(self):
        self.__instance = None
        self.args_present = None
        self.args_in_fct = None
        # self.args_required = None
        self.features = None
        self.service = self.__get_service()

    def __call__(self, **kwargs):

        if not self.__instance:
            # Update self.attrs
            self.args_present = list(kwargs.keys())
            # self.features = [x for x in dir(self.service) if
            #                  not x.startswith("__") and hasattr(getattr(self.service, x), "__call__")]
            self.features = [f[0] for f in inspect.getmembers(self.service, inspect.isfunction) if not f[0].startswith('_')]

            # Define available args for init
            sig = inspect.signature(self.initialize)
            self.args_in_fct = [x.name for x in sig.parameters.values()]
            # self.args_required = [x.name for x in sig.parameters.values() if x.empty == inspect.Parameter.empty]
            params = dict([x for x in kwargs.items() if x[0] in self.args_in_fct])

            # Initialize service
            config = self.initialize(**params)
            if config and self.service:
                self.__instance = self.service(config)
            else:
                raise RuntimeError("Builder initialisation failure, check method and arguments.")

        return self.__instance

    @classmethod
    def __get_service(cls):
        if cls.service:
            return cls.service
        else:
            raise AttributeError("Builder class must have set 'service' attribute")

    @staticmethod
    def initialize(**params):
            return params
        # raise NotImplementedError("Builder class must implement 'initialize' method")


class CryptoService:
    """
    Template to design a **Crypto Service**. Called by an associated builder,
    it must implement abstraction methods from the ``interface`` module
    (encrypt & decrypt).

    By default, ``__init__`` method will provide the **cipher** (config) built
    by the Crypto builder. It can be personalised to fulfill the needs of the
    implemented cryptography recipe.

    All defined Crypto Services must inherit from the ``CryptoService`` class.

    Example::

        class MyService(CryptoService):

            def encrypt(self, data):
                ...
                enc = self.cipher.encrypt(data)
                return enc

            def decrypt(self, data):
                ...
                plain = self.cipher.decrypt(data)
                return plain

    Note:
        See this **warning** on :ref:`sensitive data in memory <memory-warning>`.
    """

    def __init__(self, cipher):
        self.cipher = cipher

    def encrypt(self, data):
        raise NotImplementedError("Service class must implement 'encrypt' method")

    def decrypt(self, data):
        raise NotImplementedError("Service class must implement 'decrypt' method")
